/*
 * Copyright (C) 2006-2013 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file dccp_send.c Paraslash's dccp sender. */

/*
 * based on server.c of dccp-cs-0.01.tar.bz2,
 * (C) 2005 Ian McDonald <imcdnzl@gmail.com>
 */

#include <netinet/in.h>
#include <sys/socket.h>
#include <regex.h>
#include <sys/types.h>
#include <osl.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <netdb.h>

#include "para.h"
#include "error.h"
#include "string.h"
#include "afh.h"
#include "afs.h"
#include "server.h"
#include "net.h"
#include "list.h"
#include "send.h"
#include "sched.h"
#include "vss.h"
#include "fd.h"
#include "close_on_fork.h"
#include "chunk_queue.h"
#include "server.cmdline.h"
#include "acl.h"

static struct sender_status dccp_sender_status, *dss = &dccp_sender_status;

struct dccp_fec_client {
	struct fec_client_parms fcp;
	struct fec_client *fc;
};

static void dccp_pre_select(int *max_fileno, fd_set *rfds,
		__a_unused fd_set *wfds)
{
	if (dss->listen_fd >= 0)
		para_fd_set(dss->listen_fd, rfds, max_fileno);
}

/**
 * Query the TX CCID used on the sender-client half connection.
 * \param sockfd Server socket descriptor to query (after accept(2)).
 * \return CCID number > 0, -1 on error/incompatibility.
 *
 * NB: This feature is only available on Linux > 2.6.30; on older kernels
 * ENOPROTOOPT ("Protocol not available") will be returned.
 */
static int dccp_get_tx_ccid(int sockfd)
{
	int tx_ccid;
	socklen_t len = sizeof(tx_ccid);

	if (getsockopt(sockfd, SOL_DCCP,
		       DCCP_SOCKOPT_TX_CCID, &tx_ccid, &len) < 0) {
		PARA_WARNING_LOG("DCCP_SOCKOPT_TX_CCID: %s\n", strerror(errno));
		return -1;
	}
	return tx_ccid;
}

static void dccp_shutdown_client(struct sender_client *sc)
{
	struct dccp_fec_client *dfc = sc->private_data;

	vss_del_fec_client(dfc->fc);
	shutdown_client(sc, dss);
}

static void dccp_shutdown_clients(void)
{
	struct sender_client *sc, *tmp;

	list_for_each_entry_safe(sc, tmp, &dss->client_list, node)
		dccp_shutdown_client(sc);
}

/** * Obtain current MPS according to RFC 4340, sec. 14. */
static int dccp_init_fec(struct sender_client *sc)
{
	int mps, ret;
	socklen_t ml = sizeof(mps);

	/* If call fails, return some sensible minimum value */
	ret = getsockopt(sc->fd, SOL_DCCP, DCCP_SOCKOPT_GET_CUR_MPS, &mps, &ml);
	if (ret < 0) {
		PARA_NOTICE_LOG("can not determine MPS: %s\n", strerror(errno));
		mps = generic_max_transport_msg_size(sc->fd) - DCCP_MAX_HEADER;
	}
	PARA_INFO_LOG("current MPS = %d bytes\n", mps);
	assert(mps > 0);
	if (conf.dccp_max_slice_size_arg > 0)
		mps = PARA_MIN(mps, conf.dccp_max_slice_size_arg);
	return mps;
}

static void dccp_send_fec(struct sender_client *sc, char *buf, size_t len)
{
	int ret = xwrite(sc->fd, buf, len);

	if (ret < 0)
		dccp_shutdown_client(sc);
}

static void dccp_post_select(fd_set *rfds, __a_unused fd_set *wfds)
{
	struct sender_client *sc;
	struct dccp_fec_client *dfc;
	int tx_ccid;

	sc = accept_sender_client(dss, rfds);
	if (!sc)
		return;

	/* If CCID identifiable, present client as <host>#<port>~<ccid> */
	tx_ccid = dccp_get_tx_ccid(sc->fd);
	if (tx_ccid != -1) {
		char *tmp = make_message("%s~%d", sc->name, tx_ccid);

		free(sc->name);
		sc->name = tmp;
	}
	/*
	 * Bypass unused CCID paths: the sender does not receive application data
	 * from the client; by shutting down this unused communication path we can
	 * reduce processing costs a bit. See analogous comment in dccp_recv.c.
	 */
	if (shutdown(sc->fd, SHUT_RD) < 0) {
		PARA_WARNING_LOG("%s\n", strerror(errno));
		shutdown_client(sc, dss);
		return;
	}
	dfc = para_calloc(sizeof(*dfc));
	sc->private_data = dfc;
	dfc->fcp.data_slices_per_group	= conf.dccp_data_slices_per_group_arg;
	dfc->fcp.slices_per_group	= conf.dccp_slices_per_group_arg;
	dfc->fcp.init_fec		= dccp_init_fec;
	dfc->fcp.send_fec		= dccp_send_fec;
	dfc->fcp.need_periodic_header   = false;
	dfc->fc = vss_add_fec_client(sc, &dfc->fcp);
}

static int dccp_com_on(__a_unused struct sender_command_data *scd)
{
	return generic_com_on(dss, IPPROTO_DCCP);
}

static int dccp_com_off(__a_unused struct sender_command_data *scd)
{
	dccp_shutdown_clients();
	generic_com_off(dss);
	return 1;
}


static int dccp_com_deny(struct sender_command_data *scd)
{
	generic_com_deny(scd, dss);
	return 1;
}

static int dccp_com_allow(struct sender_command_data *scd)
{
	generic_com_allow(scd, dss);
	return 1;
}

/**
 * Return list of available CCIDs or warning, in static buffer.
 */
static const char *dccp_list_available_ccids(void)
{
	/* Worst case length: n * 3 digits + n-1 spaces + '\0' */
	static char list[DCCP_MAX_HOST_CCIDS * 4];
	uint8_t *ccids;
	int i, len, nccids;

	nccids = dccp_available_ccids(&ccids);
	if (nccids < 0) {
		snprintf(list, sizeof(list), "Unable to query available CCIDs");
	} else {
		for (i = len = 0; i < nccids; i++)
			len += snprintf(list + len, sizeof(list) - len,
					"%s%d", i ? " " : "", ccids[i]);
	}
	return list;
}

static char *dccp_info(void)
{
	char *info = get_sender_info(dss, "dccp");
	char *ret  = make_message("%s" "\tsupported ccids: %s\n",
				  info, dccp_list_available_ccids());
	free(info);
	return ret;
}

/**
 * The init function of the dccp sender.
 *
 * \param s pointer to the dccp sender struct.
 *
 * It initializes all function pointers of \a s and starts
 * listening on the given port.
 */
void dccp_send_init(struct sender *s)
{
	int ret, k, n;

	s->info = dccp_info;
	s->send = NULL;
	s->pre_select = dccp_pre_select;
	s->post_select = dccp_post_select;
	s->shutdown_clients = dccp_shutdown_clients;
	s->resolve_target = NULL;
	s->help = generic_sender_help;
	s->client_cmds[SENDER_ON] = dccp_com_on;
	s->client_cmds[SENDER_OFF] = dccp_com_off;
	s->client_cmds[SENDER_DENY] = dccp_com_deny;
	s->client_cmds[SENDER_ALLOW] = dccp_com_allow;
	s->client_cmds[SENDER_ADD] = NULL;
	s->client_cmds[SENDER_DELETE] = NULL;

	k = conf.dccp_data_slices_per_group_arg;
	n = conf.dccp_slices_per_group_arg;

	if (k <= 0 || n <= 0 || k >= n) {
		PARA_WARNING_LOG("invalid FEC parameters, using defaults\n");
		conf.dccp_data_slices_per_group_arg = 3;
		conf.dccp_slices_per_group_arg = 4;
	}

	init_sender_status(dss, conf.dccp_access_arg, conf.dccp_access_given,
		conf.dccp_port_arg, conf.dccp_max_clients_arg,
		conf.dccp_default_deny_given);
	ret = generic_com_on(dss, IPPROTO_DCCP);
	if (ret < 0)
		PARA_ERROR_LOG("%s\n", para_strerror(-ret));
}
