/*
 * Copyright (C) 2005-2013 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file udp_send.c Para_server's udp sender. */

#include <netinet/in.h>
#include <sys/socket.h>
#include <regex.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/udp.h>
#include <net/if.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <osl.h>
#include <netdb.h>

#include "server.cmdline.h"
#include "para.h"
#include "error.h"
#include "string.h"
#include "afh.h"
#include "afs.h"
#include "server.h"
#include "list.h"
#include "send.h"
#include "sched.h"
#include "vss.h"
#include "portable_io.h"
#include "net.h"
#include "fd.h"
#include "close_on_fork.h"

/**
 * Time window during which ICMP Destination/Port Unreachable messages are
 * ignored, covering transient receiver problems such as restarting the
 * client, rebooting, reconfiguration, or handover.
 */
#define UDP_MAX_UNREACHABLE_TIME 30

/** Describes one entry in the list of targets for the udp sender. */
struct udp_target {
	/** Track time (seconds) of last ICMP Port Unreachable error */
	time_t last_unreachable;
	/** The opaque structure returned by vss_add_fec_client(). */
	struct fec_client *fc;
	/** The FEC parameters for this target. */
	struct fec_client_parms fcp;
};

static struct list_head targets;
static int sender_status;

static void udp_close_target(struct sender_client *sc)
{
	const char *buf;
	size_t len = vss_get_fec_eof_packet(&buf);

	/*
	 * Ignore the return value of wirte() since we are closing the target
	 * anyway. The sole purpose of the "do_nothing" statement is to silence
	 * gcc.
	 */
	if (write(sc->fd, buf, len))
		do_nothing;
}

static void udp_delete_target(struct sender_client *sc, const char *msg)
{
	struct udp_target *ut = sc->private_data;

	PARA_NOTICE_LOG("deleting %s (%s) from list\n", sc->name, msg);
	udp_close_target(sc);
	close(sc->fd);
	del_close_on_fork_list(sc->fd);
	vss_del_fec_client(ut->fc);
	list_del(&sc->node);
	free(sc->name);
	free(sc);
	free(ut);
}

/**
 * Perform AF-independent multicast sender setup.
 *
 * \param sc	The connected sender client.
 *
 * \return Zero if okay, negative on error.
 */
static int mcast_sender_setup(struct sender_client *sc)
{
	struct sockaddr_storage ss;
	socklen_t sslen = sizeof(ss);
	int ttl = conf.udp_ttl_arg, id = 0;
	const int on = 1;

	if (conf.udp_mcast_iface_given) {
		char *iface = conf.udp_mcast_iface_arg;

		id = if_nametoindex(iface);
		if (id == 0)
			PARA_WARNING_LOG("could not resolve interface '%s', "
					"using default", iface);
	}

	if (getpeername(sc->fd, (struct sockaddr *)&ss, &sslen) < 0)
		goto err;
	assert(ss.ss_family == AF_INET || ss.ss_family == AF_INET6);

	/* RFC 3493, 5.2: -1 means 'use kernel default' */
	if (ttl < 0 || ttl > 255)
		ttl = -1;

	switch (ss.ss_family) {
	case AF_INET:
		if (!IN_MULTICAST(htonl(((struct sockaddr_in *)&ss)->sin_addr.s_addr)))
			return 0;
		if (id != 0) {
#ifdef HAVE_IP_MREQN
			struct ip_mreqn mn;

			memset(&mn, 0, sizeof(mn));
			mn.imr_ifindex = id;
			if (setsockopt(sc->fd, IPPROTO_IP, IP_MULTICAST_IF, &mn, sizeof(mn)) < 0)
				goto err;
#else
			PARA_ERROR_LOG("No support for setting outgoing IPv4 mcast interface.");
#endif
		}
		/*
		 * Enable receiving multicast messages generated on the local host
		 * At least on Linux, this is enabled by default.
		 */
		if (setsockopt(sc->fd, IPPROTO_IP, IP_MULTICAST_LOOP, &on, sizeof(on)) < 0)
			break;

		/* Default: use local subnet (do not flood out into the WAN) */
		if (ttl == -1)
			ttl = 1;
		if (setsockopt(sc->fd, IPPROTO_IP, IP_MULTICAST_TTL, &ttl, sizeof(ttl)) < 0)
			break;
		return 0;
	case AF_INET6:
		if (!IN6_IS_ADDR_MULTICAST(&((struct sockaddr_in6 *)&ss)->sin6_addr))
			return 0;
		if (id != 0 &&
		    setsockopt(sc->fd, IPPROTO_IPV6, IPV6_MULTICAST_IF, &id, sizeof(id)) < 0)
			break;
		if (setsockopt(sc->fd, IPPROTO_IPV6, IPV6_MULTICAST_LOOP, &on, sizeof(on)) < 0)
			break;
		if (setsockopt(sc->fd, IPPROTO_IPV6, IPV6_MULTICAST_HOPS, &ttl, sizeof(ttl)) < 0)
			break;
		return 0;
	}
err:
	return -ERRNO_TO_PARA_ERROR(errno);
}

static void udp_init_session(struct sender_client *sc)
{
	PARA_NOTICE_LOG("sending to udp %s\n", sc->name);
}

static void udp_shutdown_targets(void)
{
	struct sender_client *sc, *tmp;
	list_for_each_entry_safe(sc, tmp, &targets, node)
		udp_close_target(sc);
}

static int udp_resolve_target(const char *url, struct sender_command_data *scd)
{
	const char *result;
	int ret, port;

	ret = parse_fec_url(url, scd);
	if (ret)
		return ret;
	port = scd->port > 0 ? scd->port : conf.udp_default_port_arg;

	ret = para_connect_simple(IPPROTO_UDP, scd->host, port);
	if (ret < 0)
		return ret;

	result = remote_name(ret);
	close(ret);

	if (!parse_url(result, scd->host, sizeof(scd->host), &scd->port))
		return -E_ADDRESS_LOOKUP;
	return 1;
}

static int udp_com_on(__a_unused struct sender_command_data *scd)
{
	sender_status = SENDER_ON;
	return 1;
}

static int udp_com_off(__a_unused struct sender_command_data *scd)
{
	udp_shutdown_targets();
	sender_status = SENDER_OFF;
	return 1;
}

static struct sender_client *udp_lookup_target(struct sender_command_data *scd)
{
	struct sender_client *sc, *tmp;
	char host[MAX_HOSTLEN];
	int32_t port;

	list_for_each_entry_safe(sc, tmp, &targets, node) {
		parse_url(sc->name, host, sizeof(host), &port);

		/* Unspecified port means wildcard port match */
		if (scd->port > 0 && scd->port != port)
			continue;
		if (strcmp(host, scd->host))
			continue;
		return sc;
	}
	return NULL;
}

static int udp_com_delete(struct sender_command_data *scd)
{
	struct sender_client *sc = udp_lookup_target(scd);

	if (sc) {
		udp_delete_target(sc, "com_delete");
		return 1;
	}
	PARA_NOTICE_LOG("not deleting non-existing target '%s'\n", scd->host);
	return -E_TARGET_NOT_FOUND;
}

/** Initialize UDP session and set maximum payload size. */
static int udp_init_fec(struct sender_client *sc)
{
	int mps;

	udp_init_session(sc);
	mps = generic_max_transport_msg_size(sc->fd) - sizeof(struct udphdr);
	PARA_INFO_LOG("current MPS = %d bytes\n", mps);
	return mps;
}

/** Check and clear socket error if any. */
static int udp_check_socket_state(struct sender_client *sc)
{
	struct udp_target *ut = sc->private_data;
	int ret;
	socklen_t errlen = sizeof(ret);

	if (getsockopt(sc->fd, SOL_SOCKET, SO_ERROR, &ret, &errlen) < 0) {
		PARA_ERROR_LOG("SO_ERROR failed: %s\n", strerror(ret));
		return 0;
	} else if (ret == 0) {
		return 0;
	} else if (ret == ECONNREFUSED) {
		time_t dist = now->tv_sec - ut->last_unreachable;

		if (dist <= UDP_MAX_UNREACHABLE_TIME) {
			return 0;
		} else if (dist > 2 * UDP_MAX_UNREACHABLE_TIME) {
			ut->last_unreachable = now->tv_sec;
			return 0;
		} else {
			/*
			 * unreachable_time < dist <= 2 * unreachable_time
			 * No errors are allowed during this time window.
			 */
			PARA_NOTICE_LOG("Evicting %s after %d seconds "
					"of connection errors.\n",
					sc->name, (int)dist);
		}
	}
	return -ERRNO_TO_PARA_ERROR(ret);
}

static void udp_send_fec(struct sender_client *sc, char *buf, size_t len)
{
	int ret;

	if (sender_status == SENDER_OFF)
		return;
	if (len == 0)
		return;
	ret = udp_check_socket_state(sc);
	if (ret < 0)
		goto fail;
	ret = xwrite(sc->fd, buf, len);
	if (ret == -ERRNO_TO_PARA_ERROR(ECONNREFUSED)) {
		/*
		 * Happens if meanwhile an ICMP Destination / Port Unreachable
		 * has arrived. Ignore, persistent errors will be caught above.
		 */
		ret = 0;
	}
	if (ret < 0)
		goto fail;
	return;
fail:
	udp_delete_target(sc, para_strerror(-ret));
}

static int udp_com_add(struct sender_command_data *scd)
{
	int ret;
	struct udp_target *ut;
	struct sender_client *sc = udp_lookup_target(scd);

	if (sc) {
		PARA_NOTICE_LOG("target %s already exists - not adding it\n",
				sc->name);
		return -E_TARGET_EXISTS;
	}
	ut = para_calloc(sizeof(*ut));
	sc = para_calloc(sizeof(*sc));
	ut->fcp.slices_per_group      = scd->slices_per_group;
	ut->fcp.data_slices_per_group = scd->data_slices_per_group;
	ut->fcp.init_fec	      = udp_init_fec;
	ut->fcp.send_fec	      = udp_send_fec;
	ut->fcp.need_periodic_header  = true;

	sc->private_data = ut;
	sc->fd = -1;
	ret = para_connect_simple(IPPROTO_UDP, scd->host, scd->port);
	if (ret < 0)
		goto err;
	sc->fd = ret;

	ret = mcast_sender_setup(sc);
	if (ret < 0)
		goto err;
	ret = mark_fd_nonblocking(sc->fd);
	if (ret < 0)
		goto err;
	sc->name = para_strdup(remote_name(sc->fd));
	PARA_INFO_LOG("adding to target list (%s)\n", sc->name);
	ut->fc = vss_add_fec_client(sc, &ut->fcp);
	para_list_add(&sc->node, &targets);
	add_close_on_fork_list(sc->fd);
	return 1;
err:
	if (sc->fd >= 0)
		close(sc->fd);
	PARA_NOTICE_LOG("failed to set up %s:%d (%s)- not adding it\n",
			scd->host, scd->port, para_strerror(-ret));
	free(sc);
	free(ut);
	return ret;
}

static char *udp_info(void)
{
	struct sender_client *sc;
	char *ret, *tgts = NULL;

	list_for_each_entry(sc, &targets, node) {
		struct udp_target *ut = sc->private_data;
		char *tmp = make_message("%s%s/%u:%u ",
			tgts ? : "", sc->name,
			ut->fcp.data_slices_per_group,
			ut->fcp.slices_per_group
		);
		free(tgts);
		tgts = tmp;
	}
	ret = make_message(
		"udp sender:\n"
		"\tstatus: %s\n"
		"\tport: %s\n"
		"\ttargets: %s\n",
		(sender_status == SENDER_ON)? "on" : "off",
		stringify_port(conf.udp_default_port_arg, "udp"),
		tgts? tgts : "(none)"
	);
	free(tgts);
	return ret;
}

static void udp_init_target_list(void)
{
	struct sender_command_data scd;
	int i;

	INIT_LIST_HEAD(&targets);
	for (i = 0; i < conf.udp_target_given; i++) {
		if (udp_resolve_target(conf.udp_target_arg[i], &scd) < 0)
			PARA_CRIT_LOG("not adding requested target '%s'\n",
					conf.udp_target_arg[i]);
		else
			udp_com_add(&scd);
	}
}

static char *udp_help(void)
{
	return make_message(
		"usage: {on|off}\n"
		"usage: {add|delete} ip_address[:port][/[packet_size:]k:n]\n"
		"       - k is the number of data slices per FEC group\n"
		"       - n is the total number of slices in a FEC group\n"
		"       - packet_size reduces the slice size below path MTU\n\n"
		"examples: add 224.0.1.38:1500  (IPv4 multicast)\n"
		"          add 224.0.1.38:8080/14:16 (explicit FEC)\n"
		"          add 10.10.1.42       (using default port)\n"
		"          add [FF05::42]:1500  (IPv6 multicast)\n"
		"          add [::1]            (IPv6 localhost/default port)\n"
		"          delete myhost.net    (host with port wildcard)\n"
		"          delete 10.1.2.3:456  (IPv4 with explicit port)\n"
		"          delete [badc0de::1]  (IPv6 with port wildcard)\n"
	);
}

/**
 * The init function of para_server's udp sender.
 *
 * \param s Pointer to the udp sender struct.
 *
 * It initializes all function pointers of \a s and the list of udp targets.
 */
void udp_send_init(struct sender *s)
{
	INIT_LIST_HEAD(&targets);
	s->info = udp_info;
	s->help = udp_help;
	s->send = NULL;
	s->pre_select = NULL;
	s->post_select = NULL;
	s->shutdown_clients = udp_shutdown_targets;
	s->resolve_target = udp_resolve_target;
	s->client_cmds[SENDER_ON] = udp_com_on;
	s->client_cmds[SENDER_OFF] = udp_com_off;
	s->client_cmds[SENDER_DENY] = NULL;
	s->client_cmds[SENDER_ALLOW] = NULL;
	s->client_cmds[SENDER_ADD] = udp_com_add;
	s->client_cmds[SENDER_DELETE] = udp_com_delete;
	sender_status = SENDER_OFF;
	udp_init_target_list();
	if (!conf.udp_no_autostart_given)
		sender_status = SENDER_ON;
	PARA_DEBUG_LOG("udp sender init complete\n");
}
